classdef Formation < handle
    % FORMATION Collects all agents and takes care of their communication
    % - Checks for distance dependent and obstacle dependent communication
    % - Takes care of the communication between the agents
    % - Asks the agents to update states and outputs
    
    properties (Access = private)
        % variables
        agentIds = [];              % vector holding the ids of all
        % agents currently present in the
        % simulation
        formationRefInput = [];     % [r_x, ..., r_P]' vector holding the current
        % absolute reference input for the
        % formation's virtual leader
        relRefInput = [];           % 3-D matrix, each z-slice holds a 2-D matrix
        % [r_1, ..., r_N]' for one agent with
        % all relative distances to the other
        % agents
        outputs = [];               % [v_1, ..., v_N]'
        coordinates = [];           % [v_1, ..., v_N]'
        adjacency = [];             % NxN adjacency matrix
        agentDistances = [];        % NxN matrix, each row holds the distances
        % to all other agents for the
        % corresponding agent
        agents = cell(0);           % cell array holding the agent objects
        % objects
        memory
    end
    
    methods 
        %% Constructor
        function this = Formation(initialNumberOfAgents,formationRefInput,...
            relRefInput,formationSpecSwitch,adjacency,models,...
            output2Coordinates)
        
        % input data
        this.formationRefInput = formationRefInput;
        this.relRefInput = relRefInput;
        this.adjacency = adjacency;
        
        % create agents
        this.agents = cell(initialNumberOfAgents + 1,1);
        this.addVirtualLeader(output2Coordinates,this.formationRefInput);
        for l = 1:initialNumberOfAgents
            this.addAgent( l+1, models{l}, output2Coordinates,this.relRefInput(:,:,l) );

            % FIXME: Remove commented
%             if strcmp(formationSpecSwitch,'directed Vectors') % if reference input = distances
%                 this.addAgent(l + 1,models{l},output2Coordinates,this.relRefInput(:,:,l));
%             else % reference input = directed vectors
%                 this.addAgent(l + 1,models{l},output2Coordinates,this.relRefInput(:,l));
%             end
        end
        this.agentsSend();
        
        % create memory
        this.memory = Memory();
        end % constructor
        
        
        %% Update communication topology depending on distance/obstacles
        function modifyAdjacency(this,useCommRangeLimit,maxCommunicationDist,...
            useObstacles,obstacles,originalAdjacency)
        this.adjacency = originalAdjacency;
        
        % distance-depending interruption of communication links:
        if useCommRangeLimit
            this.calculateAgentDistances();
            N = size(this.adjacency,1);
            for l = 1:N-1
                for m = l+1:N
                    if l~=m % do not check with itself
                        if this.agentDistances(l,m) > maxCommunicationDist % agents far apart
                            % check form l to m
                            if originalAdjacency(l,m) > 0 % nagative numbers are unbreakable
                                this.adjacency(l,m) = 0;
                            end
                            % check from m to l
                            if originalAdjacency(m,l) > 0 % nagative numbers are unbreakable
                                this.adjacency(m,l) = 0;
                            end
                        end
                    end % if
                end % for
            end % for
        end % if
        
        % obstacle-depending interruption of communication links:
        if useObstacles
            for m = 1:length(this.adjacency) % for each communication link:
                for n = (m+1):length(this.adjacency)
                    if this.adjacency(m,n) == 1 || this.adjacency(n,m) == 1     % link between m and n
                        A = this.coordinates(m,:);  % first agent
                        B = this.coordinates(n,:);  % second agent
                        AB = (B-A);
                        ABn = AB/norm(AB+eps)^2;        % add eps to avoid problem at A==B
                        for l = 1:length(obstacles)
                            if this.adjacency(m,n) == 1 || this.adjacency(n,m) == 1
                                inbetween = false;
                                switch lower(obstacles{l}.type)
                                    case 'ball'
                                        % second check, as it might be that one obstacle already
                                        % broke the communication between them
                                        P = obstacles{l}.center;    % center of a ball obstacle
                                        Proj = A + AB*(P-A)'*ABn;     % projection of the center on the line A-B
                                        if norm(P-Proj) < obstacles{l}.radius    % obstacle in between
                                            inbetween = true;
                                        end
                                    case 'wall'
                                        t = (obstacles{l}.corner-A)/[AB; -obstacles{l}.d1; -obstacles{l}.d2];
                                        if all(t>0 & t<1) || any(isnan(t))
                                            inbetween = true;
                                        end
                                end % types of obstacles
                                if inbetween    % obstacle between the agents
                                    if this.adjacency(m,n) == 1
                                        this.adjacency(m,n) = 0;
                                    end
                                    if this.adjacency(n,m) == 1
                                        this.adjacency(n,m) = 0;
                                    end
                                end
                            else
                                break % the loop over agents
                            end
                        end
                    end
                end % inner loop over agents
            end % outer loop over agents
        end % if
        
        end % modifyAdjacency
        
        
        %% Communication and calculation functions   
        function calculateStep(this,timedelaySwitch,timedelay,t_s,...
            formationSpecSwitch)
        this.agentsReceiveFromOthers(timedelaySwitch,timedelay,t_s);
        this.agentsReceiveRefInput(formationSpecSwitch); % maybe only after refinput change!!!!
        for l = 1:length(this.agentIds)
            this.agents{l}.calculateStep(formationSpecSwitch);
        end
        this.agentsSend();
        end % calculateStep
        
        function agentsSend(this)
        % receive output errors from all agents:
        for l = 1:length(this.agentIds)
            this.outputs(l,:) = this.agents{l}.sendOutput();
            this.coordinates(l,:) = this.agents{l}.getCoordinates();
        end
        end % agentsSend        
        
        
        function agentsReceiveFromOthers(this,timedelaySwitch,timedelay,t_s)
        % send output errors to all agents, depending on the
        % communication topology and the time delay:
        %numberOfReceivableAgents = sum(this.adjacency,2); %column vector with number of receivable agents for each agent
        for l = 2:length(this.agentIds)
            switch timedelaySwitch.Type
                case 'none'
                    receivableOutputs = this.outputs;
                    ownSavedOutput = this.outputs(l,:)';
                    for m = 1:length(this.agentIds)
                        % if the agent cannot receive from the current agent,
                        % delete the vector entry:
                        if this.adjacency(l,m) == 0
                            receivableOutputs(m,:) = NaN;
                        end
                    end
                case 'fixed'
                    switch timedelaySwitch.DelayUnits
                        case 'steps', delay_steps = timedelay;
                        case 'sec.',  delay_steps = (ceil(timedelay/t_s));
                    end
                    receivableOutputs = NaN(size(this.outputs));
                    for m = 1:length(this.agentIds)
                        if this.adjacency(l,m) % ~= 0
                            receivableOutputs(m,:) = this.memory.getBackDatedOutput(...
                                delay_steps,this.agents{m}.getId())';
                        end
                    end
                    if timedelaySwitch.DelayOwn
                        ownSavedOutput = this.memory.getBackDatedOutput(...
                            delay_steps,this.agents{l}.getId());
                    else
                        ownSavedOutput = this.outputs(l,:)';
                    end
                case 'distanceDependend'
            end %switch
            
            this.agents{l}.receiveOutputOthers(receivableOutputs,...
                ownSavedOutput);
        end %for
        end % agentsReceiveFromOthers

        
        %% Reference signal function
        function agentsReceiveRefInput(this,formationSpecSwitch)
        % send absolute formation reference input to the virtual leader
        this.agents{1}.receiveRefInput(this.formationRefInput);
        % send relative reference input to all agents
        for l = 2:length(this.agentIds)
            if strcmp(formationSpecSwitch,'directed Vectors')
                this.agents{l}.receiveRefInput(this.relRefInput(:,:,l - 1));
            else this.agents{l}.receiveRefInput(this.relRefInput(:,l - 1));
            end % if
        end
        end % agentsReceiveRefInput        

        %% Assisting functions
        function addAgent(this,id,model,output2Coordinates,relRefInput)
        newAgent = Agent(id,model,output2Coordinates,relRefInput);
        this.agents{length(this.agentIds) + 1} = newAgent;
        this.outputs(length(this.agentIds) + 1,:) = this.agents{length(this.agentIds) + 1}.sendOutput()';
        this.coordinates(length(this.agentIds) + 1,:) = this.agents{length(this.agentIds) + 1}.getCoordinates()';
        this.agentIds = [this.agentIds id];
        end % addAgent
        
        function addVirtualLeader(this,output2Coordinates,refInput)
        newAgent = VirtualLeader(1,output2Coordinates,refInput); %id=0
        this.agents{1} = newAgent;
        this.outputs(1,:) = this.agents{1}.sendOutput()';
        this.agentIds = 1;  % that's always agent 0
        end % addVirtualLeader        
        
        function removeAgent(this,numberOfAgent,formationSpecSwitch)
        remove_mask = (this.agentIds == numberOfAgent);
        this.agents(remove_mask) = [];
        this.outputs(remove_mask,:) = [];
        this.coordinates(remove_mask,:) = [];
        this.agentIds(remove_mask) = [];
        if strcmp(formationSpecSwitch,'directed Vectors')
            this.relRefInput(:,:,remove_mask) = [];
        else this.relRefInput(:,remove_mask) = [];
        end % if
        end % removeAgent
        
        function calculateAgentDistances(this)
        N = length(this.agentIds);
        this.agentDistances = zeros(N,N);
        for l = 1:N-1
            for m = (l+1):N
                this.agentDistances(l,m) = norm(this.coordinates(l,:) - this.coordinates(m,:));
            end
        end
        this.agentDistances = this.agentDistances' + this.agentDistances;
        end % calculateAgentDistances

        
        
        %% Data recording
        function saveAgentData(this,timestep)
        this.memory.saveOutputs(this,timestep);
        end % saveAgentData
        
        function saveAdjacency(this,timestep)
        this.memory.saveAdjacency(this.adjacency,timestep);
        end % saveAdjacency
        
        
        %% Get methods        
        function outputs = getOutputs(this)
        outputs = this.outputs;
        end % getOutputs
        
        function coordinates = getCoordinates(this)
        coordinates = this.coordinates;
        end % getCoordinates
        
        function numberOfAgents = getNumberOfAgents(this)
        numberOfAgents = length(this.agentIds);
        end % getNumberOfAgents
        
        function agentIds = getAgentIds(this)
        agentIds = this.agentIds;
        end % getAgentIds
        
        function agentId = getAgentId(this,numberOfAgent)
        agentId = this.agents{numberOfAgent}.getId();
        end % getAgentId
        
        function agent = getAgent(this,numberOfAgent)
        agent = this.agents{numberOfAgent};
        end % getAgent
        
        function agentOutputs = getAgentOutputs(this,agentId,DoF)
        agentOutputs = this.memory.getOutputs(agentId,DoF);
        end % getAgentOutputs
        
        function agentTimesteps = getAgentTimesteps(this,agentId)
        agentTimesteps = this.memory.getTimesteps(agentId);
        end % getAgentTimesteps
        
        function agentData = getAgentData(this)
        agentData = this.memory.getAgentStorage();
        end % getAgentData
        
        function adjacencyData = getAdjacencyData(this)
        adjacencyData = this.memory.getAdjacencyStorage();
        end % getAdjacencyData
        
        function adjacency = getAdjacency(this)
        adjacency = this.adjacency;
        end % getAdjacency        
        
        
        %% Set methods        
        function setAdjacency(this,newAdjacency)
        this.adjacency = newAdjacency;
        end % setAdjacency
        
        function setRelRefInput(this,newRelRefInput)
        this.relRefInput = newRelRefInput;
        end % setRelRefInput
        
        function setFormationRefInput(this,newFormationRefInput)
        this.formationRefInput = newFormationRefInput;
        end % setFormationRefInput
    end
end

