classdef LinearModel < Model
    % LinearModel LTI model and controller of an agent
    % 
    % To construct such an object use
    %   LinearModel( K, P, x0, y0, v0, u0 )
    % where
    % P  - model of the agent,   discrete-time, state-space model
    % K  - controller,           discrete-time, state-space model
    % x0 - initial states of P
    % y0 - initial output of P send to the outer agents
    % v0 - initial states of K
    % u0 - initial output of K  (currently not used!)
    %
    % The fields v0 and u0 are optional - if missing or empty they are
    % assigned zero value
    
    % 4 Feb 2010, Andrey Popov    
    
    properties (Access = private)
        controllerModel             % state space model
                                    % v(k+1)=A*v(k)+B*y_cl(k)
                                    %   u(k)=C*v(k)+D*y_cl(k)
        plantModel                  % state space model
                                    % x(k+1)=A*x(k)+B*u(k)
                                    %   y(k)=C*x(k)
        outputDim = int32(0);        % determines the first n entries of the
                                    % plant output vector, which are
                                    % transmitted to the other agents. The
                                    % remaining entries of the vector are
                                    % fed back to the controller to close
                                    % the inner loop
    end
    
    methods
        % FIXME: remove function this = LinearModel(controllerModel,plantModel,x_0,y_0)
        function this = LinearModel(plantModel,controllerModel,x0,y0,v0,u0)
        this = this@Model();
        this.plantModel = plantModel;
        this.controllerModel = controllerModel;
        if nargin < 6 || isempty(u0)
            u0 = zeros(size(controllerModel.C,1),1);
        end
        if nargin < 5 || isempty(v0)
            v0 = zeros(size(controllerModel.A,1),1);
        end
        this.x_kplus1 = x0;
        this.v_kplus1 = v0;
        this.y_k = zeros(size(plantModel.C,1),1);
        this.y_0 = y0;  % initial position - used as off-set
        this.u_k = u0;
        this.outputDim = length(y0);
        end % constructor
        
        function y_0 = getInitialOutput(this)
        y_0 = this.y_0;
        end % getInitialOutput
        
        function t_s = getSamplingTime(this)
        t_s = this.plantModel.Ts;
        end % getSamplingTime
        
        function output = calculateStep(this,e_k)
        this.u_k = this.controllerModel.C*this.v_kplus1 + ...
            this.controllerModel.D*[e_k; this.y_k(this.outputDim + 1:end)];
        this.y_k = this.plantModel.C*this.x_kplus1;
        this.v_kplus1 = this.controllerModel.A*this.v_kplus1 +...
            this.controllerModel.B*[e_k; this.y_k(this.outputDim + 1:end)];
        this.x_kplus1 = this.plantModel.A*this.x_kplus1 +...
            this.plantModel.B*this.u_k;
        output = this.y_k(1:this.outputDim) + this.y_0;
        end % calculateStep
    end
    
end

