classdef SimulationManager < handle
    % SIMULATIONMANAGER Takes care of executing the simulation step in the
    % correct order
    
    
    properties (Access = private)
        %variables
        t_s = double(1);                %sampling time
        settings = struct()             %simulation settings
        idCount = int32(0);             %holds the hightest agent id which
                                        %is or has been present in the si-
                                        %mulation
        %objects
        refInputGenerator               %reference input generator
        formation                       %formation object holding the 
                                        %formation
        animator                        %animator object
    end
    
    methods
        function this = SimulationManager(settings)
            this.settings = settings;
            
            % initial number of agents, not including virtual leader:
            this.settings.initialization.initialNumberOfAgents = ...
                size(settings.initialization.adjacency,1) - 1;
            
            % calculate relative reference input from absolute reference
            % input:
            this.refInputGenerator = RefInputGenerator();
            if strcmp(this.settings.formation.formationSpecSwitch,'directed Vectors')
                this.settings.initialization.relRefInput = ...
                    this.refInputGenerator.calculateRelRefInputVectors(...
                    settings.initialization.absRefInput);
            else this.settings.initialization.relRefInput = ...
                this.refInputGenerator.calculateRelRefInputDistance(...
                settings.initialization.absRefInput,...
                settings.formation.formationRefInput.refPos(:,1));
            end % if
            
            % sampling time:
            this.t_s = this.settings.initialization.models{1}.getSamplingTime();
            
            % id count:
            this.idCount = this.settings.initialization.initialNumberOfAgents + 1;
            
            %simulationSettings = struct(...
            %'interrupt',false,...
            %'finish',false);
            %animationSettings = struct(...
            %'showCommLinksSwitch',true,...
            %'showAgentIdSwitch',true,...
            %'showTrailsSwitch',true,...
            %'trailLength',10,...
            %'animationRate',int32(10),...
            %'axesDimSwitch','auto',... % 'auto', 'manual'
            %'axesDimensions',[]); 
            %formationSettings = struct(...
            %'controllerModel',...
            %'plantModel',...
            %'output2Coordinates',...
            %'adjacency',...
            %'x0',...
            %'y0',...
            %'formationRefInput',[],...
            %'timedelaySwitch','none',... % 'none', 'constant'
            %'timedelay',0.0,...
            %'useCommRangeLimit',false, ... %true/false
            %'maxCommunicationDist',5,...
            %'relRefInput',[],...
            %'useObstacles',false,... % true/false
            %'obstacles',[]);
            %this.settings.simulation = simulationSettings;
            %this.settings.formation = formationSettings;
            %this.settings.animation = animationSettings;
        end % constructor
        
        
        function initialization(this)
            %this.loadInitializationData();
            this.formation = Formation( ...
                this.settings.initialization.initialNumberOfAgents, ...
                this.settings.formation.formationRefInput.refPos(:,1), ...
                this.settings.initialization.relRefInput, ...
                this.settings.formation.formationSpecSwitch, ...
                this.settings.initialization.adjacency, ...
                this.settings.initialization.models,...
                this.settings.initialization.output2Coordinates);
            if size(this.settings.initialization.output2Coordinates,1) == 3
                this.animator = Animator3D(this.formation.getCoordinates());
            else this.animator = Animator2D(this.formation.getCoordinates());
            end
            this.animator.animate(...
                this.formation.getCoordinates(),...
                this.formation.getAdjacency(),...
                0,... % time 0
                this.settings.animation.showCommLinksSwitch,...
                this.settings.animation.showTrailsSwitch,...
                this.settings.animation.trailLength,...
                this.settings.animation.showAgentIdSwitch,...
                this.formation.getAgentIds(),...
                this.settings.animation.axesDimSwitch,...
                this.settings.animation.axesDimensions,...
                this.settings.formation.useObstacles,...
                this.settings.formation.obstacles);
            
            this.saveAgentData(0);
            this.saveAdjacency(0)
        end % initialization
        
        function interrupt = getInterrupt(this)
            interrupt = this.settings.simulation.interrupt;
        end % getInterrupt
        
        function setFinish(this,finish)
            this.settings.simulation.finish = finish;
        end % setFinish
        
        function setInterrupt(this,interrupt)
            this.settings.simulation.interrupt = interrupt;
        end % setInterrupt
        
        function setFormationSettings(this,formationSettings)
            this.settings.formation = formationSettings;
            %if this.settings.timedelay < this.t_s
            %    this.settings.timedelay = this.t_s;
            %end
        end % setFormationSettings
        
        function setChangeSettings(this,changeSettings)
            this.settings.changes = changeSettings;
        end % setChangeSettings
        
        function setAnimationSettings(this,animationSettings)
            this.settings.animation = animationSettings;
        end
        
        function setInitializationSettings(this,initializationSettings)
            this.settings.initialization = initializationSettings;
        end % setInitializationSettings
        
        function simulate(this)
            k = 1;
            while this.settings.simulation.finish ~= true 
                
                % change formation reference input:
                this.changeFormationRefInput(k);
                
                % change formation constellation:
                if strcmpi(this.settings.changes.formationChange,'from file')
                    this.changeFormation(k);
                end % if
                
                % change adjacency matrix due to limited communication range
                % and obstacles:
                this.formation.modifyAdjacency(this.settings.formation.useCommRangeLimit,...
                    this.settings.formation.maxCommunicationDist,...
                    this.settings.formation.useObstacles,...
                    this.settings.formation.obstacles,...
                    this.settings.initialization.adjacency);
                
                % calculate step
                this.formation.calculateStep(this.settings.formation.timedelaySwitch,...
                    this.settings.formation.timedelay,this.t_s,...
                    this.settings.formation.formationSpecSwitch);
                
                % animate every nth timestep:
                if rem(k,this.settings.animation.animationRate) == 0
                    this.animator.animate(...
                        this.formation.getCoordinates(),...
                        this.formation.getAdjacency(),...
                        [num2str(this.t_s*k) ' sec.    ' num2str(k) ' steps'],...
                        this.settings.animation.showCommLinksSwitch,...
                        this.settings.animation.showTrailsSwitch,...
                        this.settings.animation.trailLength,...
                        this.settings.animation.showAgentIdSwitch,...
                        this.formation.getAgentIds(),...
                        this.settings.animation.axesDimSwitch,...
                        this.settings.animation.axesDimensions,...
                        this.settings.formation.useObstacles,...
                        this.settings.formation.obstacles);
                end
                this.saveAgentData(k);
                this.saveAdjacency(k)
                while this.settings.simulation.interrupt == true
                    pause(0.1);
                end
                k = k + 1;
            end
        end % simulate
        
        
        function plot(this,plotChoice)
            colorVector = 'ymcrgbk';
            %try
                if ~isempty(plotChoice) && plotChoice(1) == 1
                    figure
                    hold on
                    colorCounter = 1;
                    title('x-y View');
                    xlabel('x');
                    ylabel('y');
                    for l = 2:this.idCount
                        plot(this.formation.getAgentOutputs(l,1),this.formation.getAgentOutputs(l,2),colorVector(colorCounter));
                        colorCounter = colorCounter + 1;
                        if colorCounter > 7
                            colorCounter = 1;
                        end % if
                    end % for
                    plotChoice(1) = [];
                end % if
                if ~isempty(plotChoice) && plotChoice(1) == 2
                    figure
                    hold on
                    colorCounter = 1;
                    title('x-z View');
                    xlabel('x');
                    ylabel('z');
                    for l = 2:this.idCount
                        plot(this.formation.getAgentOutputs(l,1),this.formation.getAgentOutputs(l,3),colorVector(colorCounter));
                        colorCounter = colorCounter + 1;
                        if colorCounter > 7
                            colorCounter = 1;
                        end % if
                    end % for
                    plotChoice(1) = [];
                end % if
                if ~isempty(plotChoice) && plotChoice(1) == 3
                    figure
                    hold on
                    colorCounter = 1;
                    title('y-z View');
                    xlabel('y');
                    ylabel('z');
                    for l = 2:this.idCount
                        plot(this.formation.getAgentOutputs(l,2),this.formation.getAgentOutputs(l,3),colorVector(colorCounter));
                        colorCounter = colorCounter + 1;
                        if colorCounter > 7
                            colorCounter = 1;
                        end % if
                    end % for
                    plotChoice(1) = [];
                end % if
                if ~isempty(plotChoice) && plotChoice(1) == 4
                    figure
                    hold on
                    colorCounter = 1;
                    title('t-x View');
                    xlabel('t');
                    ylabel('x');
                    for l = 2:this.idCount
                        plot(this.formation.getAgentTimesteps(l),this.formation.getAgentOutputs(l,1),colorVector(colorCounter));
                        colorCounter = colorCounter + 1;
                        if colorCounter > 7
                            colorCounter = 1;
                        end % if
                    end % for
                    plotChoice(1) = [];
                end % if
                if ~isempty(plotChoice) && plotChoice(1) == 5
                    figure
                    hold on
                    colorCounter = 1;
                    title('t-y View');
                    xlabel('t');
                    ylabel('y');
                    for l = 2:this.idCount
                        plot(this.formation.getAgentTimesteps(l),this.formation.getAgentOutputs(l,2),colorVector(colorCounter));
                        colorCounter = colorCounter + 1;
                        if colorCounter > 7
                            colorCounter = 1;
                        end % if
                    end % for
                    plotChoice(1) = [];
                end % if
                if ~isempty(plotChoice) && plotChoice(1) == 6
                    figure
                    hold on
                    colorCounter = 1;
                    title('t-z View');
                    xlabel('t');
                    ylabel('z');
                    for l = 2:this.idCount
                        plot(this.formation.getAgentTimesteps(l),this.formation.getAgentOutputs(l,3),colorVector(colorCounter));
                        colorCounter = colorCounter + 1;
                        if colorCounter > 7
                            colorCounter = 1;
                        end % if
                    end % for
                end % if
            %catch me
            %end %try
        end % plot
        
        function saveData(this,pathname,filename)
            agentData = this.formation.getAgentData();
            adjacencyData = this.formation.getAdjacencyData();
            samplingTime = this.t_s;
            save(strcat(pathname,'\',filename),'agentData','adjacencyData',...
                'samplingTime')
        end % saveData
    end
    
    %% Methods
    methods (Access = private)
        %% Update reference point coordinates
        function changeFormationRefInput(this,k)
        % Check for time-stamp
        if ~isfield(this.settings.formation.formationRefInput, 't_step')
            nsamples = size(this.settings.formation.formationRefInput.refPos,2) - 1;
            if nsamples
                this.settings.formation.formationRefInput.t_step = k + (0:nsamples-1);
            else % single position                
                this.settings.formation.formationRefInput.t_step = 0;
            end
        end
        old_times = (this.settings.formation.formationRefInput.t_step <= k);
        [time,ind] = max(this.settings.formation.formationRefInput.t_step(old_times));
        this.formation.setFormationRefInput(this.settings.formation.formationRefInput.refPos(:,ind));
        end % changeFormationRefInput
        
        %% Change formation
        function changeFormation(this,k)
            changeData = this.settings.changes.changeData;  % shorter!
            change_done = false;
            while ~isempty(changeData) && (changeData{1}.t_change <= k)
                switch changeData{1}.type
                    case 'commTopology'
                        this.settings.initialization.adjacency = ...
                            changeData{1}.adjacency;
                    case 'constellation'
                        if strcmpi(this.settings.formation.formationSpecSwitch,'directed Vectors')
                            this.settings.initialization.absRefInput = changeData{1}.absRefInput;
                            this.settings.initialization.relRefInput = ...
                                this.refInputGenerator.calculateRelRefInputVectors(...
                                changeData{1}.absRefInput);
                        else
                        end
                        this.formation.setRelRefInput(this.settings.initialization.relRefInput);
                    
                    case 'deleteAgent'
                        for l = 1:length(changeData{1}.deleteIds)
                            agentIds = this.formation.getAgentIds();    % list of existing agents
                            % note, ID=1, is the reference point
                            remove_agent = find(changeData{1}.deleteIds(l)+1 == agentIds, 1, 'first');
                            if ~isempty( remove_agent )     % there is such agent
                                m = agentIds(remove_agent);
                                % changeData{1}.deleteIds(l) + 1;
                                this.formation.removeAgent(m,...
                                    this.settings.formation.formationSpecSwitch);
                                this.settings.initialization.adjacency(m,:) = [];
                                this.settings.initialization.adjacency(:,m) = [];
                                this.settings.initialization.absRefInput(m,:) = [];
                                if strcmpi(this.settings.formation.formationSpecSwitch,'directed Vectors')
                                    this.settings.initialization.relRefInput = ...
                                        this.refInputGenerator.calculateRelRefInputVectors(...
                                        this.settings.initialization.absRefInput);
                                else
                                end
                                this.formation.setAdjacency(this.settings. ...
                                    initialization.adjacency);
                                this.formation.setRelRefInput(this.settings. ...
                                    initialization.relRefInput);
                            end % if ~isempty
                        end % for
                        
                    case 'addAgent'
                        % update reference signals
                        this.settings.initialization.absRefInput = [...
                            this.settings.initialization.absRefInput;...
                            changeData{1}.newAbsRefInput];
                        if strcmpi(this.settings.formation.formationSpecSwitch,'directed Vectors')
                            this.settings.initialization.relRefInput = ...
                                this.refInputGenerator.calculateRelRefInputVectors(...
                                this.settings.initialization.absRefInput);
                        else this.settings.initialization.relRefInput = ...
                                this.refInputGenerator.calculateRelRefInputDistance(...
                                this.settings.initialization.absRefInput,...
                                this.settings.formation.formationRefInput.refPos(:,1));
                        end % if
                        this.settings.initialization.adjacency = ...
                            changeData{1}.adjacency;
                        
                        for l = 1:length(changeData{1}.models)
                            this.formation.addAgent(this.idCount + 1,...
                                changeData{1}.models{l},...
                                this.settings.initialization.output2Coordinates,[]);
                                this.idCount = this.idCount + 1;
                        end % for
                        
                        this.formation.setAdjacency(this.settings.initialization.adjacency);
                        this.formation.setRelRefInput(this.settings.initialization.relRefInput);
                        
                end % switch
                
                % delete "used" change data:
                changeData(1) = [];
                change_done = true;
            end % while
            if change_done
                this.settings.changes.changeData = changeData;
            end
        end % changeFormation

        function saveAgentData(this,timestep)
            this.formation.saveAgentData(timestep);
        end % saveAgentData

        function saveAdjacency(this,timestep)
            this.formation.saveAdjacency(timestep);
        end % saveAdjacency
    end
    
end

